# Мета-програмиране

```elixir
# elixir/kernel.ex
defmacro defmacro(call, expr \\ nil) do
  define(:defmacro, call, expr, __CALLER__)
end
```



- Можем автоматично да дефинираме функции:

```elixir
defmodule QuickMathz do
  @values [{:one, 1}, {two: two}, {:three, 3}]

  for {name, value} <- @values do
    def unquote(name)(), do: unquote(value)
  end
end

QuickMathz.three # => 3
```



```elixir
html do
  head do
    title do
      text "Hello To Our HTML DSL"
    end
  end
  body do
    h1 class: "title" do
      text "Introduction to metaprogramming"
    end
    p do
      text "Metaprogramming with Elixir is really awesome!"
    end
  end
end
```



Ecto (Database Access, Data Mapping & Validation):

```elixir
from o in Order,
where: o.created_at > ^Timex.shift(DateTime.utc_now(), days: -2)
join: i in OrderItems, on: i.order_id == o.id
```



Plug (~~http library~~ web specification)

```elixir
get "/hello" do
  send_resp(conn, 200, "world")
end

match _ do
  send_resp(conn, 404, "oops")
end
```




Absinthe (GraphQL Server implementation)

```elixir
field :subscribe, :subscription_plan do
  arg(:plan_id, non_null(:integer))
  arg(:card_token, :string, default_value: nil)
  arg(:coupon, :string, default_value: nil)

  middleware(JWTAuth)

  resolve(&BillingResolver.subscribe/3)
end
```



ExUnit

```elixir
defmodule Blogit.ComponentTest do
  use ExUnit.Case, async: true

  describe "when a module uses it with `use Blogit.Component`" do
    test "injects a function base_name/0, which returns the name of " <>
           "the module in underscore case" do
      assert TestComponent.base_name() == "test_component"
    end
  end

  # ... блогът има повече от 1 тест, вярвайте!
end
```



В Elixir можем само така:

```elixir
assert true
assert 5 == 4
assert 5 < 4
```



#### quote

- Връща абстрактното синтактично дърво на всеки израз

```elixir
quote do: 1 + 1
# => {:+, [context: Elixir, import: Kernel], [1, 1]}
```



```elixir
quote do: sum(1, 2, 3)
# => {:sum, [], [1, 2, 3]}
```



- Пример с map, всъщност той също е представен като извъкване към `%{}`

```elixir
quote do: %{1 => 2}
# => {:%{}, [], [{1, 2}]}
```



Малко по-дълъг пример:

```elixir
quote do
  html do
    head do
      title do
        text "Hello To Our HTML DSL"
      end
    end
  end
end
```



```elixir
{:html, [],
 [
   [
     do: {:head, [],
      [
        [
          do: {:title, [],
           [
             [
               do: {:text, [],
                ["Hello to our HTML DSL"]}
             ]
           ]}
        ]
      ]}
   ]
 ]}
```



- Пример с променлива:
```elixir
quote do: x
# => {:x, [], Elixir}
```



Това са единствените неща, чийто AST е самият израз.

```elixir
:sum         #=> Atoms
1.0          #=> Numbers
[1, 2]       #=> Lists
"strings"    #=> Strings
{key, value} #=> Tuples with two elements
```



```elixir
iex(3)> x = 1
1
iex(4)> quote do: x + 1
{:+, [context: Elixir, import: Kernel], [{:x, [], Elixir}, 1]}
#                                          ^ Защо!?
```



```elixir
quote do: unquote(x) + 1
# => {:+, [context: Elixir, import: Kernel], [1, 1]}
```



- Пример с по-голям `unquote`:

```elixir
ast1 = quote do: 1 + 2
ast2 = quote do: unquote(ast1) + 3
Macro.to_string(ast2)
# => "1 + 2 + 3"
```



- Можем да сме яки и да `unquote`-ваме извикване на функция:

```elixir
fun = :hello
Macro.to_string(quote do: unquote(fun)(:world))
# => "hello(:world)"
```



- Какво става ако имаме израз от типа: `[1, 2, 6]` и искаме да вмъкнем `[3, 4, 5]` между `2` и `6`?

```elixir
inner = [3, 4, 5]
Macro.to_string(quote do: [1, 2, unquote(inner), 6])
# => "[1, 2, [3, 4, 5], 6]"
```



#### unquote_splicing

- Позволява ни да интерполираме изрази в списъци, речници и наредени n-орки


```elixir
inner = [3, 4, 5]
Macro.to_string(quote do: [1, 2, unquote_splicing(inner), 6])
# => "[1, 2, 3, 4, 5, 6]"
```



- Пример с речник:
```elixir
kw = [foo: :bar, baz: :quix]
Macro.to_string(quote do: %{unquote_splicing(kw)})
# => "%{foo: :bar, baz: :quix}"
```



- Това е опция на `quote`, която ни позволява това:

```elixir
defmodule Hello do
  defmacro say(name) do
    quote bind_quoted: [name: name] do
      "Здравей #{name}, как е?"
    end
  end
end
```



- Вместо:

```elixir
defmodule Hello do
  defmacro say(name) do
    quote do
      "Здравей #{unquote(name)}, как е?"
    end
  end
end
```



- И ето примера в действие:

```elixir
iex(1)> Hello.say("Ники")
"Здравей Ники, как е?"
iex(2)> name
** (CompileError) iex:4: undefined function name/0
```



```elixir
defmodule FMI do
  defmacro if(condition, do: do_clause, else: else_clause) do
    quote do
      case unquote(condition) do
        x when x, [false, nil] -> unquote(else_clause)
        _ -> unquote(do_clause)
      end
    end
  end
end
```




```elixir
iex(1)> require FMI
# => FMI
iex(2)> FMI.if true do
...(2)>   1
...(2)> else
...(2)>   2
...(2)> end
# => 1
```


- Всъщност if-ът ни прие като аргументи:

```elixir
FMI.if(true, [do: 1, else: 2])
```



- Краткия синтаксис за дефиниране на функция е страничен ефект от
това, че като последен аргумент може да изпускаме скобите
- `def foo(), do: :work`
- Ако добавим липсващите кръгли и квадратни скоби, получаваме:
- С други думи `do/end` блоковете са синтактична захар над keyword
списъците

```elixir
defmodule T do
  def(foo(), [do: :work])
end
```



- Всъщност `def/defp/defmacro/defmacrop/defmodule` са също макроси
- Това ни позволява да пишем ето така код:

```elixir
defmodule(Math, [
  {:do, def(add(a, b), [{:do, a + b}])}
])
```



#### Macro.expand_once

- Можем на една стъпка да "разгъваме/оценяваме" AST-та

```elixir
ast = quote do
  FMI.unless true do
    IO.puts "Hello"
  end
end
# {{:., [], [{:__aliases__, [alias: false], [:FMI]}, :unless]}, [],
#  [
#    true,
#    [
#      do: {{:., [], [{:__aliases__, [alias: false], [:IO]}, :puts]}, [],
#       ["Hello"]}
#    ]
#  ]}
```



След това:

```elixir
require FMI
Macro.expand_once(ast, __ENV__)
# {:if, [context: FMI, import: Kernel],
#  [
#    {:!, [context: FMI, import: Kernel], [true]},
#    [
#      do: {{:., [],
#        [
#          {:__aliases__, [alias: false, counter: -576460752303423390], [:IO]},
#          :puts
#        ]}, [], ["Hello"]},
#      else: nil
#    ]
#  ]}
```



- На кратко:

```elixir
defmodule OurModule do
  defmacro __using__(opts) do
    quote do
       def get_opts(), do: unquote(opts)
    end
  end
end

defmodule SeeUsing do
  use OurModule, option: "Hello"
end

SeeUsing.get_opts()
# => [option: "Hello"]
```



- Е същото като:

```elixir
defmodule SeeUsing do
  require OurModule
  OurModule.__using__(option: "Hello")
end

SeeUsing.get_opts()
# => [option: "Hello"]
```



```elixir
defmodule A do
  defmacro __before_compile__(_env) do
    quote do
      def hello, do: "world"
    end
  end
end

defmodule B do
  @before_compile A
end

B.hello()
#=> "world"
```



Пример:

```elixir
ast = quote do
  if a == 42 do
    "The answer is?"
  else
    "Mehhh"
  end
end
Code.eval_quoted ast, a: 42
# warning: variable "a" does not exist and is being expanded to "a()", please use parentheses to remove the ambiguity or chang
# e the variable name
#   nofile:1
# 
# ** (CompileError) nofile:1: undefined function a/0
#     (stdlib) lists.erl:1354: :lists.mapfoldl/3
#     (elixir) expanding macro: Kernel.if/2
#     nofile:1: (file)
# # BOOOOOOOOOM
```



```elixir
ast = quote do
  if var!(a) == 42 do
    "The answer is?"
  else
    "Mehhh"
  end
end
Code.eval_quoted ast, a: 42
# => {"The answer is?", [a: 42]}
Code.eval_quoted ast, a: 1
# => {"Mehhh", [a: 1]}
```



- Нека видим по-oпасен пример:

```elixir
defmodule Dangerous do
  defmacro rename(new_name) do
    quote do
      var!(name) = unquote(new_name)
    end
  end
end
# => {:module, Dangerous, .....
require Dangerous
# => Dangerous
name = "Слави"
# => "Слави"
Dangerous.rename("Вало")
# => "Вало"
name
# => "Вало"
```



```elixir
require Dangerous
# => Dangerous
Dangerous.rename("Вало")
# => "Вало"
name
# => "Вало"
```



- Макро, което ни дава да имаме повече от 255 аргумента.
  * Hint - какво би било ast-то на:

```elixir
quote do: sum all 1, -1
```